#!/usr/bin/perl
use lib "./UPnP/lib";
use UPnP::ControlPoint;
use Socket;
use IO::Select;
use IO::Handle;
use Data::Dumper;
use HTML::Parser;
use HTML::Entities;
use URI::Escape;
use XML::Simple;
use HTTP::Daemon;
use HTML::Template;
use LWP::MediaTypes qw(add_type);
use POSIX qw(strftime);
use strict;

###############################################################################
#CONFIG
$main::MAX_LOG_LEVEL  = 3;    # Lower, the less output
$main::HTTP_PORT      = 8001; # Port our fake http server listens on
$main::MAX_SEARCH     = 200;  # Max search results to return
$main::RENEW_SUB_TIME = 3600; # How often do we do a UPnP renew in seconds
$main::VERSION        = "0.12";
$main::DEFAULT        = "index.html";
#$main::DEFAULT        = "simple/index.html";
$| = 1;


use SOAP::Lite maptype => {}; 

$SIG{INT} = "main::quit";
$SIG{PIPE} = sub {};
$SIG{CHLD} = sub {};

$Data::Dumper::Indent = 1;
@main::CALLBACKS     = ();
@main::WAITINGAV     = ();
@main::WAITINGRENDER = ();
@main::WAITINGANY    = ();
$main::UPDATENUM      = time();

main();

###############################################################################
# main
sub main {
    Log (0, "Starting up!");
    Log (0, "If this is the last line you see, you may need to disable your firewall, or allow this application.");


    add_type("text/css" => qw(css));

    $main::daemon = HTTP::Daemon->new(LocalPort => $main::HTTP_PORT, Reuse => 1) || die;
    $main::cp = UPnP::ControlPoint->new ();
    my $search = $main::cp->searchByType("urn:schemas-upnp-org:device:ZonePlayer:1", \&main::search_cb);

    my @selsockets = $main::cp->sockets();
    @selsockets = (@selsockets, $main::daemon);
    my $select = IO::Select->new(@selsockets);


    if ( -f "musicdb.pl") {
        Log (0, "Loading Music DB");
        require "musicdb.pl";
        $main::LASTUPDATE  = $main::UPDATENUM;
        $main::MUSICUPDATE = $main::UPDATENUM++;
        Log (0, "Finished Loading Music DB");
    } else {
        Log (0, "No Music DB - Will auto ReIndex after 10 seconds");
        add_timeout (time()+10, \&reindex);
    }
    renew_subscriptions();

    # MAIN LOOP
    while (1) {

        # Check the callbacks we have waiting
        my $timeout = 5;
        my $now = time;
        while ($#main::CALLBACKS >= 0) {
            if ($main::CALLBACKS[0][0] <= $now) {
                my($time, $callback) = @{shift @main::CALLBACKS};
                &$callback;
            } else {
                $timeout = $main::CALLBACKS[0][0] - $now;
                last;
            }
        }

        # Find if any sockets are ready for reading
        my @sockets = $select->can_read($timeout);

        # Call the handlers for the sockets
        for my $sock (@sockets) {
            if ($sock == $main::daemon) {
                my $c = $main::daemon->accept;
                my $r = $c->get_request;
                handle_http_request($c, $r);
            } else {
                $main::cp->handleOnce($sock);
            }
        }
    }
}
###############################################################################
sub add_timeout {
    my ($time, $cb) = @_;

    @main::CALLBACKS = sort { $a->[0] <=> $b->[0] } @main::CALLBACKS, [$time, $cb];
}
###############################################################################
sub process_waiting_http {
    my ($list) = @_;

    foreach my $data (@{$list}) {
        Log (3, "Sending response to $data->[2]");
        send_http_response($data->[0], $data->[1], $data->[2]);
    }
    @{$list} = ();
}
###############################################################################
sub Log {
    my $level = shift;
    my $now = strftime "%e/%H%M%S", localtime;
    my @Level = ("crit", "notice", "info", "debug");

    return if ($level > $main::MAX_LOG_LEVEL);

    my ($package, $filename, $line) = caller(0);
    my ($x, $y, $z, $subroutine)    = caller(1);
    $subroutine = "*unknown*" if (!defined $subroutine);
    print "$now $Level[$level]: $filename:$line $subroutine @_\n";
}

###############################################################################
sub handle_http_request {
    my ($c, $r) = @_;

    # No r, just return
    if (!$r || !$r->uri) {
        Log (1, "Missing Request");
        return;
    }

    my $uri = $r->uri;

    my $path = $uri->path;
    my $baseurl;

    if ($r->header("host")) {
        $baseurl = "http://".$r->header("host");
    } else {
        $baseurl = "http://". UPnP::Common::LOCAL_IP.":".$main::HTTP_PORT;
    }

    if (($path eq "/") || ($path =~ /\.\./)) {
        $c->send_redirect("$baseurl/$main::DEFAULT");
        $c->force_last_request;
        return;
    }

    if (-d "html" . $path) {
        if ($path =~ /\/$/) {
            $c->send_redirect($baseurl . $path . "index.html");
        } else {
            $c->send_redirect($baseurl . $path . "/index.html");
        }
        $c->force_last_request;
        return;
    }

    my %qf = $uri->query_form;
    Log (1, "URI: $uri");

    # File doesn't exist
    if ( ! -f "html/$path") {
        $c->send_error(HTTP::Status::RC_FORBIDDEN);
        $c->force_last_request;
        return;
    }

    # File isn't HTML
    if (!($path =~ /\.html/) && !($path =~ /\.xml/)) {
        $c->send_file_response("html/$path");
        $c->force_last_request;
        return;
    }

    # 0 - not handled, 1 - handled send reply, 2 - handled routine will send reply
    my $response = 0;

    if ($qf{action}) {
        if ($qf{zone}) {
            $response = handle_http_zone_action($c, $r, $path);
        }
        if (! $response) {
            $response = handle_http_action($c, $r, $path);
        }
    }

    if ($qf{NoWait}) {
        $response = 1;
    }

    if ($response == 2) {
        push (@main::WAITINGAV, [$c, $r, $path]);
    } elsif ($response == 3) {
        push (@main::WAITINGRENDER, [$c, $r, $path]);
    } elsif ($response == 4) {
        push (@main::WAITINGANY, [$c, $r, $path]);
    } else {
        send_http_response($c, $r, $path);
    }
}

###############################################################################
sub handle_http_zone_action {
my ($c, $r, $path) = @_;

    my %qf = $r->uri->query_form;

    my $location = $main::ZONES{$qf{zone}}->{Location};
    if ($qf{action} eq "Remove") {
        avtransport_remove_track($location, $qf{queue});
        return 2;
    } elsif ($qf{action} eq "RemoveAll") {
        avtransport_action($location, "RemoveAllTracksFromQueue");
        return 2;
    } elsif ($qf{action} eq "Play") {
        avtransport_play($location);
        return 2;
    } elsif ($qf{action} =~ /(Next|Previous|Stop|Pause)/) {
        avtransport_action($location, $qf{action});
        return 2;
    } elsif ($qf{action} eq "ShuffleOn") {
        avtransport_shuffle($location, 1);
        return 2;
    } elsif ($qf{action} eq "ShuffleOff") {
        avtransport_shuffle($location, 0);
        return 2;
    } elsif ($qf{action} eq "RepeatOn") {
        avtransport_repeat($location, 1);
        return 2;
    } elsif ($qf{action} eq "RepeatOff") {
        avtransport_repeat($location, 0);
        return 2;
    } elsif ($qf{action} eq "Seek") {
        if (! ($main::AV{"$qf{zone},AVTransportURI"} =~ /queue/)) {
            avtransport_set_queue($location);
        }
        avtransport_seek($location, $qf{queue});
        return 2;
    } elsif ($qf{action} eq "MuteOn") {
        render_mute($location, 1);
        return 3;
    } elsif ($qf{action} eq "MuteOff") {
        render_mute($location, 0);
        return 3;
    } elsif ($qf{action} eq "MuchSofter") {
        render_volume_change($location, -5);
        return 3;
    } elsif ($qf{action} eq "Softer") {
        render_volume_change($location, -1);
        return 3;
    } elsif ($qf{action} eq "Louder") {
        render_volume_change($location, +1);
        return 3;
    } elsif ($qf{action} eq "MuchLouder") {
        render_volume_change($location, +5);
        return 3;
    } elsif ($qf{action} eq "Save") {
        avtransport_save($location, $qf{savename});
        return 0;
    } elsif ($qf{action} eq "AddMusic") {
        if (music_class($qf{mpath}) eq "object.item.audioItem.audioBroadcast") {
            avtransport_set_radio($location, $qf{mpath});
        } else {
            avtransport_add($location, $qf{mpath});
        }
        return 2;
    } elsif ($qf{action} eq "DeleteMusic") {
        my @parts = split("/", $qf{mpath});
        if ($#parts == 2 && music_class($qf{mpath}) eq "object.container.playlist") {
            my $id = $main::MUSIC{$parts[1]}{$parts[2]}{"id"};
            content_dir_delete($location, $id);
        }
        return 0;
    } elsif ($qf{action} eq "PlayMusic") {
        if (music_class($qf{mpath}) eq "object.item.audioItem.audioBroadcast") {
            avtransport_set_radio($location, $qf{mpath});
        } else {
            if (! ($main::AV{"$qf{zone},AVTransportURI"} =~ /queue/)) {
                avtransport_set_queue($location);
            }
            avtransport_action($location, "RemoveAllTracksFromQueue");
            avtransport_add($location, $qf{mpath});
            avtransport_play($location);
        }

        return 2;
    } else {
        return 0;
    }
    return 1;
}

###############################################################################
sub handle_http_action {
my ($c, $r, $path) = @_;

    my %qf = $r->uri->query_form;

    if ($qf{action} eq "ReIndex") {
        reindex();
    } elsif ($qf{action} eq "Wait" && $qf{lastupdate}) {
        if ($main::LASTUPDATE > $qf{lastupdate}) {
            return 1;
        } else {
            return 4;
        }
    } else {
        return 0;
    }
    return 1;
}
###############################################################################
sub get_zone_data {
my ($zone, $updatenum) = @_;

    my %activedata;

    $activedata{"ACTIVE_ZONE"}      = encode_entities($main::ZONES{$zone}->{ZoneName});
    $activedata{"ACTIVE_ZONEID"}    = uri_escape($zone); 
    $activedata{"ACTIVE_VOLUME"}    = $main::RENDER{"$zone,Volume"}->{Master}->{val};

    my $lastupdate;
    if ($main::RENDER{"$zone,LASTUPDATE"} > $main::AV{"$zone,LASTUPDATE"}) {
        $lastupdate = $main::RENDER{"$zone,LASTUPDATE"};
    } else {
        $lastupdate = $main::AV{"$zone,LASTUPDATE"};
    }

    $activedata{"ACTIVE_LASTUPDATE"} = $lastupdate;
    $activedata{"ACTIVE_UPDATED"}    = ($lastupdate > $updatenum);

    if ($main::RENDER{"$zone,Mute"}->{Master}->{val}) {
        $activedata{"ACTIVE_MUTED"} = 1;
    } else {
        $activedata{"ACTIVE_MUTED"} = 0;
    }

    my $curtrack     = $main::AV{"$zone,CurrentTrackMetaData"};
    my $curtransport = $main::AV{"$zone,AVTransportURIMetaData"};

    $activedata{"ACTIVE_NAME"}      = "";
    $activedata{"ACTIVE_ARTIST"}    = "";
    $activedata{"ACTIVE_ALBUM"}     = "";
    $activedata{"ACTIVE_ISSONG"}    = 1;
    $activedata{"ACTIVE_ISRADIO"}   = 0;
    $activedata{"ACTIVE_TRACK_NUM"} = 0;
    $activedata{"ACTIVE_TRACK_TOT"} = 0;
    $activedata{"ACTIVE_MODE"}      = 0;
    $activedata{"ACTIVE_PAUSED"}    = 0;
    $activedata{"ACTIVE_STOPPED"}   = 0;
    $activedata{"ACTIVE_PLAYING"}   = 0;
    $activedata{"ACTIVE_SHUFFLE"}   = 0;
    $activedata{"ACTIVE_REPEAT"}    = 0;

    if ($curtrack) {
        if ($curtransport && $curtransport->{item}->{"upnp:class"} eq "object.item.audioItem.audioBroadcast") {
            if (!ref($curtrack->{item}->{"r:streamContent"})) {
                $activedata{"ACTIVE_NAME"}  = encode_entities($curtrack->{item}->{"r:streamContent"});
            }

            $activedata{"ACTIVE_ALBUM"}     = encode_entities($curtransport->{item}->{"dc:title"});
            $activedata{"ACTIVE_ISSONG"}    = 0;
            $activedata{"ACTIVE_ISRADIO"}   = 1;
            $activedata{"ACTIVE_TRACK_NUM"} = 0;
            $activedata{"ACTIVE_TRACK_TOT"} = 0;
        } else {

            $activedata{"ACTIVE_NAME"}      = encode_entities($curtrack->{item}->{"dc:title"});
            $activedata{"ACTIVE_ARTIST"}    = encode_entities($curtrack->{item}->{"dc:creator"});
            $activedata{"ACTIVE_ALBUM"}     = encode_entities($curtrack->{item}->{"upnp:album"});
            $activedata{"ACTIVE_TRACK_NUM"} = $main::AV{"$zone,CurrentTrack"};
            $activedata{"ACTIVE_TRACK_TOT"} = $main::AV{"$zone,NumberOfTracks"};
        }

        if ($main::AV{"$zone,TransportState"} eq "PAUSED_PLAYBACK") {
            $activedata{"ACTIVE_MODE"}      = 2;
            $activedata{"ACTIVE_PAUSED"}    = 1;
        } elsif ($main::AV{"$zone,TransportState"} eq "STOPPED"){
            $activedata{"ACTIVE_MODE"}      = 0;
            $activedata{"ACTIVE_STOPPED"}   = 1;
        } else {
            $activedata{"ACTIVE_MODE"}      = 1;
            $activedata{"ACTIVE_PLAYING"}   = 1;
        }

        if ($main::AV{"$zone,CurrentPlayMode"} eq "NORMAL") {
        } elsif ($main::AV{"$zone,CurrentPlayMode"} eq "REPEAT_ALL") {
            $activedata{"ACTIVE_REPEAT"}    = 1;
        } elsif ($main::AV{"$zone,CurrentPlayMode"} eq "SHUFFLE_NOREPEAT") {
            $activedata{"ACTIVE_SHUFFLE"}   = 1;
        } elsif ($main::AV{"$zone,CurrentPlayMode"} eq "SHUFFLE") {
            $activedata{"ACTIVE_SHUFFLE"}   = 1;
            $activedata{"ACTIVE_REPEAT"}    = 1;
        }
    }

    my $nexttrack = $main::AV{"$zone,r:NextTrackMetaData"};
    if ($nexttrack) {
        $activedata{"NEXT_NAME"}      = encode_entities($nexttrack->{item}->{"dc:title"});
        $activedata{"NEXT_ARTIST"}    = encode_entities($nexttrack->{item}->{"dc:creator"});
        $activedata{"NEXT_ALBUM"}     = encode_entities($nexttrack->{item}->{"upnp:album"});
        $activedata{"NEXT_ISSONG"}    = 1;
    }

    return %activedata;
}

###############################################################################
sub get_queue_data {
my ($zone, $updatenum) = @_;

    my %queuedata;

    $queuedata{QUEUE_ZONE}       = encode_entities($main::ZONES{$zone}->{ZoneName});
    $queuedata{QUEUE_ZONEID}     = uri_escape($zone);
    $queuedata{QUEUE_LASTUPDATE} = $main::QUEUEUPDATE{$zone};
    $queuedata{QUEUE_UPDATED}    = ($main::QUEUEUPDATE{$zone} > $updatenum);

    return %queuedata if (!$main::QUEUE{$zone} || ! (@{$main::QUEUE{$zone}}[0]));

    my @loop_data = ();
    foreach my $queue (@{@{$main::QUEUE{$zone}}[0]}) {
        my %row_data;
        if ($main::AV{"$zone,CurrentTrackMetaData"} ne "" && 
            $queue->{res}->{content} eq $main::AV{"$zone,CurrentTrackMetaData"}->{item}->{res}->{content}) {
            if ($main::AV{"$zone,TransportState"} eq "PLAYING") {
                $row_data{QUEUE_IMG} = "/playing.gif";
            } else {
                $row_data{QUEUE_IMG} = "/paused.gif";
            }
        } else {
            $row_data{QUEUE_IMG} = "/blank.gif";
        }
        $row_data{QUEUE_NAME} = $queue->{"dc:title"};
        $row_data{QUEUE_ARG}  = "zone=" . uri_escape($zone). "&queue=$queue->{id}";
        $row_data{QUEUE_ID}   = $queue->{id};
        push(@loop_data, \%row_data);
    }
    $queuedata{QUEUE_LOOP}       = \@loop_data;

    return %queuedata;
}

###############################################################################
sub send_http_response {
my ($c, $r, $path) = @_;

    my %qf = $r->uri->query_form;

    my $updatenum = 0;

    $updatenum = $qf{lastupdate} if ($qf{lastupdate});

    # One of ours templates, now fill in the parts we know
    my $template = HTML::Template->new(filename => "html/$path",
                                       die_on_bad_params => 0,
                                       cache => 1,
                                       loop_context_vars => 1);

    $template->param("VERSION"     => $main::VERSION);
    $template->param("LAST_UPDATE" => $main::LASTUPDATE);

    if ($r->header("host")) {
        $template->param("BASE_URL" => "http://".$r->header("host"));
    } else {
        $template->param("BASE_URL" => "http://". UPnP::Common::LOCAL_IP.":".$main::HTTP_PORT);
    }

    # There is a ZONES_LOOP so gather all the information about the zones
    if ($template->query(name => "ZONES_LOOP")) {
        $template->param("ZONES_LASTUPDATE" => $main::ZONESUPDATE);
        $template->param("ZONES_UPDATED" => ($main::ZONESUPDATE > $updatenum));
        my @loop_data = ();
        foreach my $zone (keys %main::ZONES) {
            my %row_data = get_zone_data($zone, $updatenum);
            $row_data{ZONE_NAME} = encode_entities($main::ZONES{$zone}->{ZoneName});
            $row_data{ZONE_ARG}  = "zone=".uri_escape($zone);
            push(@loop_data, \%row_data);
        }
        $template->param(ZONES_LOOP => \@loop_data);
    }
    
    # There was a zone param, get the data about single zone
    if ($qf{zone}) {
        my %activedata = get_zone_data($qf{zone}, $updatenum);
        $template->param(%activedata);
    }

    my $all_arg = $r->uri->query;
    $all_arg =~ s/[&]*action=[^&]+//;
    $all_arg =~ s/[&]*rand=[^&]+//;
    $all_arg =~ s/[&]*mpath=[^&]+//;
    $template->param("ALL_ARG" => "$all_arg&rand=$$.$main::COUNTER");
    $main::COUNTER++;

    if ($template->query(name => "ALL_QUEUE_LOOP")) {
        my @loop_data = ();
        foreach my $zone (keys %main::ZONES) {
            my %row_data = get_queue_data($zone, $updatenum);

            push(@loop_data, \%row_data);
        }
        $template->param(ALL_QUEUE_LOOP => \@loop_data);
    }

    if ($template->query(name => "QUEUE_LOOP") && 
        $main::QUEUE{$qf{zone}} && 
        @{$main::QUEUE{$qf{zone}}}[0]) {

        my %queuedata = get_queue_data($qf{zone}, $updatenum);
        $template->param(%queuedata);
    }

    if ($template->query(name => "MUSIC_LOOP") && ($main::QUEUE{$qf{zone}} || $qf{lastupdate})) {
        my @loop_data = ();

        my $sortsub = sub {};

        if ($qf{msearch}) {
            foreach my $artist (sort {$a cmp $b} keys %main::MUSIC) {
                if ($artist =~ /$qf{msearch}/i) {
                    my %row_data;
                    $row_data{MUSIC_NAME} = encode_entities($artist);
                    $row_data{MUSIC_PATH} = "/" . uri_escape($artist, "^A-Za-z0-9");
                    $row_data{MUSIC_ARG} = "zone=" . uri_escape($qf{zone}) . 
                                          "&mpath=" . $row_data{MUSIC_PATH};
                    $row_data{MUSIC_ISSONG} = 0;
                    push(@loop_data, \%row_data);
                }
                if ($#loop_data > $main::MAX_SEARCH) {
                    $template->param("MUSIC_ERROR" => "More then $main::MAX_SEARCH matching items.<BR>");
                    last;
                }
                foreach my $album (sort {$a cmp $b} keys %{$main::MUSIC{$artist}}) {
                    if ($album =~ /$qf{msearch}/i) {
                        my %row_data;
                        $row_data{MUSIC_NAME} = encode_entities($album);
                        $row_data{MUSIC_PATH} = "/" . uri_escape($artist, "^A-Za-z0-9") .
                                                "/" . uri_escape($album, "^A-Za-z0-9") ;
                        $row_data{MUSIC_ARG} = "zone=" . uri_escape($qf{zone}) . 
                                                       "&mpath=" . $row_data{MUSIC_PATH};
                        if ($artist eq "#Playlists") {
                            $row_data{MUSIC_ISSONG} = 1;
                        } else {
                            $row_data{MUSIC_ISSONG} = 0;
                        }
                        push(@loop_data, \%row_data);
                    }
                    if ($#loop_data > $main::MAX_SEARCH) {
                        $template->param("MUSIC_ERROR" => "More then $main::MAX_SEARCH matching items.<BR>");
                        last;
                    }
                    foreach my $song (sort {$a cmp $b} keys %{$main::MUSIC{$artist}{$album}}) {
                        if ($song =~ /$qf{msearch}/i) {
                            my %row_data;
                            $row_data{MUSIC_NAME} = encode_entities($song);
                            $row_data{MUSIC_PATH} = "/" . uri_escape($artist, "^A-Za-z0-9") .
                                                    "/" . uri_escape($album, "^A-Za-z0-9") .
                                                    "/" . uri_escape($song, "^A-Za-z0-9");
                            $row_data{MUSIC_ARG} = "zone=" . uri_escape($qf{zone}) . 
                                                  "&mpath=".$row_data{MUSIC_PATH};
                            $row_data{MUSIC_ISSONG} = 1;
                            push(@loop_data, \%row_data);
                        }
                    }
                }
            }
            $template->param("MUSIC_LOOP" => \@loop_data);
            $template->param("MUSIC_PATH" => "Search: ". uri_escape($qf{msearch}));
            $template->param("MUSIC_SEARCH" => $qf{msearch});
            $template->param("MUSIC_UPDATED" => 1);
            $template->param("MUSIC_LASTUPDATE" => $main::MUSICUPDATE);

        } else {
            my $mpath = join("&",$qf{mpath});
            my $mhash;
            my @parts = split("/", $mpath);

            if (($#parts > 1) && ($parts[1] eq "#Playlists")) {
                @parts = ("", "#Playlists");
                $mpath = "";
            }

            if ($#parts == 1) {
                $mhash = \%{$main::MUSIC{$parts[1]}};
                $sortsub = sub {$a cmp $b};
                $template->param("MUSIC_UPDATED" => 1);
                $mpath = "/".uri_escape($parts[1], "^A-Za-z0-9");
            } elsif ($#parts == 2) {
                $mhash = \%{$main::MUSIC{$parts[1]}{$parts[2]}};
                $sortsub = sub {$mhash->{$a}{"upnp:originalTrackNumber"} <=> $mhash->{$b}{"upnp:originalTrackNumber"}};
                $template->param("MUSIC_UPDATED" => 1);
                $mpath = "/".uri_escape($parts[1], "^A-Za-z0-9").
                         "/".uri_escape($parts[2], "^A-Za-z0-9");
            } else {
                $template->param("MUSIC_ROOT" => "1");
                $sortsub = sub {$a cmp $b};
                $mhash = \%main::MUSIC;
                $mpath = "";
                $template->param("MUSIC_UPDATED" => ($main::MUSICUPDATE > $updatenum));
            }
            $template->param("MUSIC_LASTUPDATE" => $main::MUSICUPDATE);
            $template->param("MUSIC_PATH" => $mpath);

            foreach my $music (sort $sortsub keys %$mhash) {
                my %row_data;
                $row_data{MUSIC_NAME} = $music;
                $row_data{MUSIC_PATH} = $mpath . "/" . uri_escape($music, "^A-Za-z0-9");
                $row_data{MUSIC_ARG} = "zone=" . uri_escape($qf{zone}) . 
                                     "&mpath=" . $row_data{MUSIC_PATH};
                if ($#parts == 2 || (($#parts == 1) && ($parts[1] eq "#Playlists"))) {
                    $row_data{MUSIC_ISSONG} = 1;
                } else {
                    $row_data{MUSIC_ISSONG} = 0;
                }
                push(@loop_data, \%row_data);
            }
            $template->param("MUSIC_LOOP" => \@loop_data);
        }
    }
    
    my $response = HTTP::Response->new(200, undef, [Connection => "close", "Content-Type" => "text/html; charset=ISO-8859-1"], $template->output);
    $c->send_response($response);
    $c->force_last_request;
    $c->close;
}

###############################################################################
sub fetch_radio {
# Unlike music library, this shouldn't be huge

    # Pick the first coordinator's location to fetch everything from
    my @zones = keys (%main::ZONES);
    my $location = $main::ZONES{$main::ZONES{$zones[0]}->{Coordinator}}->{Location};

    undef $main::MUSIC{"#Radio"};

    my $folders =  content_dir_browse($location, "R:");
    return if (!defined $folders->[0]);

    foreach my $folder (@{$folders->[0]}) {
        my $stations =  content_dir_browse($location, $folder->{"id"});
        if ($stations && @{$stations}[0]) {
            foreach my $station (@{@{$stations}[0]}) {
                $main::MUSIC{"#Radio"}{$folder->{"dc:title"}}{$station->{"dc:title"}} = $station;
            }
            $main::LASTUPDATE  = $main::UPDATENUM;
            $main::MUSICUPDATE = $main::UPDATENUM++;
        }
    }

    process_waiting_http(\@main::WAITINGANY);
}

###############################################################################
sub fetch_playlists {
# Unlike music library, this shouldn't be huge

    # Pick the first coordinator's location to fetch everything from
    my @zones = keys (%main::ZONES);
    my $location = $main::ZONES{$main::ZONES{$zones[0]}->{Coordinator}}->{Location};

    undef $main::MUSIC{"#Playlists"};

    my $playlists =  content_dir_browse($location, "SQ:");
    return if (!defined $playlists->[0]);
    foreach my $playlist (@{$playlists->[0]}) {
        $main::MUSIC{"#Playlists"}{$playlist->{"dc:title"}} = $playlist;
    }
    $main::LASTUPDATE  = $main::UPDATENUM;
    $main::MUSICUPDATE = $main::UPDATENUM++;

    process_waiting_http(\@main::WAITINGANY);
}

###############################################################################
sub reindex {

    # Pick the first coordinator's location to fetch everything from
    my @zones = keys (%main::ZONES);
    my $location = $main::ZONES{$main::ZONES{$zones[0]}->{Coordinator}}->{Location};

    my $contentDir = get_service($main::DEVICE{$location}, "urn:schemas-upnp-org:service:ContentDirectory:1");
    $main::REINDEX{contentDirProxy} = $contentDir->controlProxy; 
    $main::REINDEX{start} = 0;
    undef %main::REINDEX_DATA;
    Log (0, "Fetching DB");
    reindex_callback();
}

###############################################################################
sub reindex_callback {
    # We only do 100 at a time otherwise we lock out everything else
    my $result = $main::REINDEX{contentDirProxy}->Browse("A:TRACKS", 'BrowseDirectChildren', 
                                          'dc:title,res,dc:creator,upnp:artist,upnp:album', 
                                          $main::REINDEX{start}, 200, "");

    $main::REINDEX{start} += $result->getValue("NumberReturned");
    Log (0, "DB: $main::REINDEX{start} of " . $result->getValue("TotalMatches"));

    my $results = $result->getValue("Result");
    my $tree = XMLin($results, forcearray => ["item"]);

    foreach my $objectid (keys %{$tree->{item}}) {
        my $entry = $tree->{item}{$objectid};
        $entry->{"dc:creator"} = "Unknown" if (! $entry->{"dc:creator"});
        $entry->{"upnp:album"} = "Unknown" if (! $entry->{"upnp:album"});

        $main::REINDEX_DATA{$entry->{"dc:creator"}}{$entry->{"upnp:album"}}{$entry->{"dc:title"}} = $entry;
        $main::REINDEX_DATA{$entry->{"dc:creator"}}{$entry->{"upnp:album"}}{$entry->{"dc:title"}}{"objectid"} = {$objectid};
    }

    if ($main::REINDEX{start} >= $result->getValue("TotalMatches")) {
        %main::MUSIC = %main::REINDEX_DATA;
        undef %main::REINDEX_DATA;

        {
            local $Data::Dump::Purity = 1;
            Log (0, "Saving Music DB");
            open (DB, ">musicdb.pl");
            my $dumper = Data::Dumper->new( [\%main::MUSIC], [ qw( *main::MUSIC) ] );
            print DB $dumper->Dump();
            close DB;
            Log (0, "Finshed Saving Music DB");
        }

        add_timeout (time(), \&fetch_radio);
    } else {
        add_timeout (time(), \&reindex_callback);
    }
}
###############################################################################
sub renew_subscriptions {
    foreach my $sub (keys %main::SUBSCRIPTIONS) {
        Log (2, "renew $sub");
        $main::SUBSCRIPTIONS{$sub}->renew;
    }
    add_timeout(time()+$main::RENEW_SUB_TIME, \&renew_subscriptions);
}

###############################################################################
sub quit {
    foreach my $sub (keys %main::SUBSCRIPTIONS) {
        Log (1, "Unsubscribe $sub");
        $main::SUBSCRIPTIONS{$sub}->unsubscribe;
    }
    $main::daemon->close();
    $main::daemon->shutdown(2);
    undef $main::daemon;
    Log (0, "Shutting Down");
    exit 0;
}

###############################################################################
sub get_service {
    my ($device, $name) = @_;

    my $service = $device->getService($name);
    return $service if ($service);

    for my $child ($device->children) {
        $service = $child->getService($name);
        return $service if ($service);
    }
    return undef;
}

###############################################################################
sub get_zone_id {
    my ($location) = @_;

    foreach my $zone (keys %main::ZONES) {
        return $zone if ($main::ZONES{$zone}->{Location} eq $location);
    }
    return undef;
}

###############################################################################
sub service_sub_cb {
    my ($service, %properties) = @_;

    my $zone = get_zone_id($service->{BASE});
    Log (1, "Event received for service=" . $service->{BASE}. " id=" . $service->serviceId . " zone=" . $zone);

    #Log(2, "Data " . Dumper(\%properties));

    # Save off the zone names
    if ($service->serviceId =~ /serviceId:ZoneGroupTopology/) {
        my $tree = XMLin(decode_entities($properties{ZoneGroupState}), 
                forcearray => ["ZoneGroup", "ZoneGroupMember"]);

        foreach my $group (@{$tree->{ZoneGroup}}) {
            my %zonegroup = %{$group};

            foreach my $member (@{$zonegroup{ZoneGroupMember}}) {
                my $zkey = $member->{UUID};
                $main::ZONES{$zkey} = $member;
                $main::ZONES{$zkey}->{Coordinator} = $zonegroup{Coordinator};
            }
        }
        $main::LASTUPDATE  = $main::UPDATENUM;
        $main::ZONESUPDATE = $main::UPDATENUM++;

        process_waiting_http(\@main::WAITINGANY);

        return;
    }

    # Save off the current status
    if ($service->serviceId =~ /serviceId:RenderingControl/) {
        my $tree = XMLin(decode_entities($properties{LastChange}), 
                forcearray => ["ZoneGroup"], 
                keyattr=>{"Volume"   => "channel",
                          "Mute"     => "channel",
                          "Loudness" => "channel"});
        foreach my $key ("Volume", "Treble", "Bass", "Mute", "Loudness") {
            if ($tree->{InstanceID}->{$key}) {
                $main::RENDER{"$zone,$key"} = $tree->{InstanceID}->{$key};
                $main::LASTUPDATE                 = $main::UPDATENUM;
                $main::RENDER{"$zone,LASTUPDATE"} = $main::UPDATENUM++;
            }
        }

        process_waiting_http(\@main::WAITINGRENDER);
        process_waiting_http(\@main::WAITINGANY);

        return;
    }

    if ($service->serviceId =~ /serviceId:AVTransport/) {
        my $tree = XMLin(decode_entities($properties{LastChange}));

        foreach my $key ("CurrentTrackMetaData", "CurrentPlayMode", "NumberOfTracks", "CurrentTrack", "TransportState", "AVTransportURIMetaData", "AVTransportURI", "r:NextTrackMetaData") {
            if ($tree->{InstanceID}->{$key}) {
                $main::LASTUPDATE             = $main::UPDATENUM;
                $main::AV{"$zone,LASTUPDATE"} = $main::UPDATENUM++;
                if ($tree->{InstanceID}->{$key}->{val} =~ /^</) {
                    $main::AV{"$zone,$key"} = XMLin($tree->{InstanceID}->{$key}->{val});
                } else {
                    $main::AV{"$zone,$key"} = $tree->{InstanceID}->{$key}->{val};
                }
            }  
        }

        process_waiting_http(\@main::WAITINGAV);
        process_waiting_http(\@main::WAITINGANY);

        return;
    }

    if ($service->serviceId =~ /serviceId:ContentDirectory/) {
        if (!defined $main::QUEUE{$zone} || $properties{ContainerUpdateIDs} =~ /Q:0/) {
            Log (1, "Refetching Q for $main::ZONES{$zone}->{ZoneName} updateid $properties{ContainerUpdateIDs}");
            $main::QUEUE{$zone} = content_dir_browse($service->{BASE}, "Q:0");
            $main::LASTUPDATE = $main::UPDATENUM;
            $main::QUEUEUPDATE{$zone} = $main::UPDATENUM++;
            process_waiting_http(\@main::WAITINGANY);
        }

        if (!defined $main::MUSIC{"#Radio"} || $properties{ContainerUpdateIDs} =~ /R:/) {
            Log (1, "Refetching Radio Stations, update id $properties{ContainerUpdateIDs}");
            add_timeout (time(), \&fetch_radio);
        }

        if (!defined $main::MUSIC{"#Playlists"} || $properties{ContainerUpdateIDs} =~ /SQ:/) {
            Log (1, "Refetching Playlists, update id $properties{ContainerUpdateIDs}");
            add_timeout (time(), \&fetch_playlists);
        }
    }


    while (my ($key, $val) = each %properties) {
        if ($val =~ /&lt/) {
            my $d = decode_entities($val);
            my $tree = XMLin($d, forcearray => ["ZoneGroup"], keyattr=>{"ZoneGroup" => "ID"});
            Log(2, "Property ${key}'s value is " . Dumper($tree));
        } else {
            Log(2, "Property ${key}'s value is " . $val);
        }
    }
}

###############################################################################
sub content_dir_browse {
    my ($location, $objectid) = @_;

    my $start = 0;
    my @data = ();
    my $result;

    my $contentDir = get_service($main::DEVICE{$location}, "urn:schemas-upnp-org:service:ContentDirectory:1");
    my $contentDirProxy = $contentDir->controlProxy; 

    do {
        $result = $contentDirProxy->Browse($objectid, 'BrowseDirectChildren', 
                                           'dc:title,res,dc:creator,upnp:artist,upnp:album', 
                                           $start, 5000, "");

        return undef if (!$result->isSuccessful);

        $start += $result->getValue("NumberReturned");

        my $results = $result->getValue("Result");

        my $tree = XMLin($results, forcearray => ["item", "container"], keyattr=>{"item" => "ID"});

        push(@data, $tree->{item}) if (defined $tree->{item});
        push(@data, $tree->{container}) if (defined $tree->{container});
    } while ($start < $result->getValue("TotalMatches"));

    return \@data;
}

###############################################################################
sub content_dir_delete {
    my ($location, $objectid) = @_;

    my $contentDir = get_service($main::DEVICE{$location}, "urn:schemas-upnp-org:service:ContentDirectory:1");
    my $contentDirProxy = $contentDir->controlProxy; 

    $contentDirProxy->DestroyObject($objectid);
}

###############################################################################
sub avtransport_remove_track {
    my ($location, $objectid) = @_;

    my $avTransport = get_service($main::DEVICE{$location}, "urn:schemas-upnp-org:service:AVTransport:1");
    my $avTransportProxy = $avTransport->controlProxy;

    my $result = $avTransportProxy->RemoveTrackFromQueue("0", $objectid);
    return;
}
###############################################################################
sub avtransport_play {
    my ($location) = @_;

    my $avTransport = get_service($main::DEVICE{$location}, "urn:schemas-upnp-org:service:AVTransport:1");
    my $avTransportProxy = $avTransport->controlProxy;

    my $result = $avTransportProxy->Play("0", "1");
    return $result;
}
###############################################################################
sub avtransport_seek {
    my ($location,$queue) = @_;

    my $avTransport = get_service($main::DEVICE{$location}, "urn:schemas-upnp-org:service:AVTransport:1");
    my $avTransportProxy = $avTransport->controlProxy;

    $queue =~ s,^.*/,,;

    my $result = $avTransportProxy->Seek("0", "TRACK_NR", $queue);
    return $result;
}
###############################################################################
sub render_mute {
    my ($location,$on) = @_;

    my $render = get_service($main::DEVICE{$location}, "urn:schemas-upnp-org:service:RenderingControl:1");
    my $renderProxy = $render->controlProxy;
    my $zone = get_zone_id($location);
    my $result = $renderProxy->SetMute("0", "Master", $on);
    $main::RENDER{"$zone,Mute"}->{Master}->{val} = $on;
    return $result;
}
###############################################################################
sub render_volume_change {
    my ($location,$change) = @_;

    my $zone = get_zone_id($location);
    my $vol = $main::RENDER{"$zone,Volume"}->{Master}->{val} + $change;
    render_volume($location,$vol);
}
###############################################################################
sub render_volume {
    my ($location,$vol) = @_;

    my $render = get_service($main::DEVICE{$location}, "urn:schemas-upnp-org:service:RenderingControl:1");
    my $renderProxy = $render->controlProxy;
    my $zone = get_zone_id($location);
    my $result = $renderProxy->SetVolume("0", "Master", $vol);
    $main::RENDER{"$zone,Volume"}->{Master}->{val} = $vol;
    return $result;
}
###############################################################################
sub avtransport_action {
    my ($location,$action) = @_;

    my $avTransport = get_service($main::DEVICE{$location}, "urn:schemas-upnp-org:service:AVTransport:1");
    my $avTransportProxy = $avTransport->controlProxy;

    my $result = $avTransportProxy->$action("0");
    return $result;
}
###############################################################################
sub avtransport_repeat {
    my ($location, $repeat) = @_;

    my $avTransport = get_service($main::DEVICE{$location}, "urn:schemas-upnp-org:service:AVTransport:1");
    my $avTransportProxy = $avTransport->controlProxy;
    my $zone = get_zone_id($location);
    my $str = $main::AV{"$zone,CurrentPlayMode"};

    if ($str eq "NORMAL") {
        $str = "REPEAT_ALL" if ($repeat);
    } elsif ($str eq "REPEAT_ALL") {
        $str = "NORMAL" if (!$repeat);
    } elsif ($str eq "SHUFFLE_NOREPEAT") {
        $str = "SHUFFLE" if ($repeat);
    } elsif ($str eq "SHUFFLE") {
        $str = "SHUFFLE_NOREPEAT" if (!$repeat);
    }
    my $result = $avTransportProxy->SetPlayMode("0", $str);
    $main::AV{"$zone,CurrentPlayMode"} = $str;
    return $result;
}
###############################################################################
sub avtransport_shuffle {
    my ($location, $shuffle) = @_;

    my $avTransport = get_service($main::DEVICE{$location}, "urn:schemas-upnp-org:service:AVTransport:1");
    my $avTransportProxy = $avTransport->controlProxy;
    my $zone = get_zone_id($location);
    my $str = $main::AV{"$zone,CurrentPlayMode"};

    if ($str eq "NORMAL") {
        $str = "SHUFFLE_NOREPEAT" if ($shuffle);
    } elsif ($str eq "REPEAT_ALL") {
        $str = "SHUFFLE" if ($shuffle);
    } elsif ($str eq "SHUFFLE_NOREPEAT") {
        $str = "NORMAL" if (!$shuffle);
    } elsif ($str eq "SHUFFLE") {
        $str = "REPEAT_ALL" if (!$shuffle);
    }

    my $result = $avTransportProxy->SetPlayMode("0", $str);
    $main::AV{"$zone,CurrentPlayMode"} = $str;
    return $result;
}
###############################################################################
sub avtransport_add {
    my ($location, $mpath) = @_;

    my @parts = split("/", $mpath);

    if ($#parts == 0) {
        foreach my $music (sort {$a cmp $b} keys %main::MUSIC) {
            avtransport_add($location, "/$music");
        }
        return;
    } elsif ($#parts == 1) {
        foreach my $music (sort {$a cmp $b} keys %{$main::MUSIC{$parts[1]}}) {
            avtransport_add($location, "$mpath/$music");
        }
        return;
    } elsif (($#parts == 2) && ($parts[1] ne "#Playlists")) {
        my $mhash = \%{$main::MUSIC{$parts[1]}{$parts[2]}};
        foreach my $music (sort {$mhash->{$a}{"upnp:originalTrackNumber"} <=> $mhash->{$b}{"upnp:originalTrackNumber"}} keys %{$mhash}) {
            avtransport_add($location, "$mpath/$music");
        }
        return;
    } 
    
    my $entry;
    if (($#parts == 2) && ($parts[1] eq "#Playlists")) {
        $entry = $main::MUSIC{$parts[1]}{$parts[2]};
    } else {
        $entry = $main::MUSIC{$parts[1]}{$parts[2]}{$parts[3]};
    }
    if ($entry->{"upnp:class"} eq "object.item.audioItem.audioBroadcast") {
        return
    } 
    my $avTransport = get_service($main::DEVICE{$location}, "urn:schemas-upnp-org:service:AVTransport:1");
    my $avTransportProxy = $avTransport->controlProxy;
    my $result = $avTransportProxy->AddURIToQueue(0, $entry->{res}->{content}, "");

    return;
}
###############################################################################
sub avtransport_set_radio {
    my ($location, $mpath) = @_;

    my @parts = split("/", $mpath);

    my $entry;
    if (($#parts == 2) && ($parts[1] eq "#Radio")) {
        $entry = $main::MUSIC{$parts[1]}{$parts[2]};
    } else {
        $entry = $main::MUSIC{$parts[1]}{$parts[2]}{$parts[3]};
    }
    my $avTransport = get_service($main::DEVICE{$location}, "urn:schemas-upnp-org:service:AVTransport:1");
    my $avTransportProxy = $avTransport->controlProxy;

# So, I'm very lazy :-)
    my $urimetadata = '&lt;DIDL-Lite xmlns:dc=&quot;http://purl.org/dc/elements/1.1/&quot; xmlns:upnp=&quot;urn:schemas-upnp-org:metadata-1-0/upnp/&quot; xmlns:r=&quot;urn:schemas-rinconnetworks-com:metadata-1-0/&quot; xmlns=&quot;urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/&quot;&gt;&lt;item id=&quot;' . $entry->{id} . '&quot; parentID=&quot;' . $entry->{parentID} . '&quot; restricted=&quot;true&quot;&gt;&lt;dc:title&gt;'. $entry->{"dc:title"} .  '&lt;/dc:title&gt;&lt;upnp:class&gt;object.item.audioItem.audioBroadcast&lt;/upnp:class&gt;&lt;desc id=&quot;cdudn&quot; nameSpace=&quot;urn:schemas-rinconnetworks-com:metadata-1-0/&quot;&gt;RINCON_AssociatedZPUDN&lt;/desc&gt;&lt;/item&gt;&lt;/DIDL-Lite&gt;';

    my $result = $avTransportProxy->SetAVTransportURI(0, $entry->{res}->{content}, decode_entities($urimetadata));

    return;
}
###############################################################################
sub avtransport_set_queue {
    my ($location) = @_;

    my $avTransport = get_service($main::DEVICE{$location}, "urn:schemas-upnp-org:service:AVTransport:1");
    my $avTransportProxy = $avTransport->controlProxy;

    my $zone = get_zone_id($location);
    my $result = $avTransportProxy->SetAVTransportURI(0, "x-rincon-queue:" . $zone . "#0", "");

    return;
}
###############################################################################
sub avtransport_save {
    my ($location, $name) = @_;

    my $avTransport = get_service($main::DEVICE{$location}, "urn:schemas-upnp-org:service:AVTransport:1");
    my $avTransportProxy = $avTransport->controlProxy;

    my $zone = get_zone_id($location);
    my $result = $avTransportProxy->SaveQueue(0, $name, "");

    return;
}
###############################################################################
sub music_class {
    my ($mpath) = @_;

    my @parts = split("/", $mpath);

    if ($#parts == 1) {
        return $main::MUSIC{$parts[1]}{"upnp:class"};
    } elsif ($#parts == 2) {
        return $main::MUSIC{$parts[1]}{$parts[2]}{"upnp:class"};
    } elsif ($#parts == 3) {
        return $main::MUSIC{$parts[1]}{$parts[2]}{$parts[3]}{"upnp:class"};
    } 

    return undef;
}
###############################################################################
sub search_cb {
    my ($search, $device, $action) = @_;
    if ($action eq 'deviceAdded') {
        Log(1, "Device: " . $device->friendlyName . " added.");
        #print Dumper($device);

        $main::DEVICE{$device->{LOCATION}} = $device;
         
        my $zonegroup = get_service($device, "urn:schemas-upnp-org:service:ZoneGroupTopology:1");
        $main::SUBSCRIPTIONS{"$device->{LOCATION},zonegroup"} = $zonegroup->subscribe(\&service_sub_cb);

        my $contentDir = get_service($device, "urn:schemas-upnp-org:service:ContentDirectory:1");
        $main::SUBSCRIPTIONS{"$device->{LOCATION},contentDir"} = $contentDir->subscribe(\&service_sub_cb);
        $main::contentDirProxy = $contentDir->controlProxy; 

        my $avtransport = get_service($device, "urn:schemas-upnp-org:service:AVTransport:1");
        $main::SUBSCRIPTIONS{"$device->{LOCATION},avtransport"} = $avtransport->subscribe(\&service_sub_cb);

        my $rendercontrol = get_service($device, "urn:schemas-upnp-org:service:RenderingControl:1");
        $main::SUBSCRIPTIONS{"$device->{LOCATION},rendercontrol"} = $rendercontrol->subscribe(\&service_sub_cb);
    }
    elsif ($action eq 'deviceRemoved') {
        Log(1, "Device: " . $device->friendlyName . " removed.");
    } else {
        Log(1, "Device: " . $device->friendlyName . " unknown action: $action.");
    }
}
